"""
This file contains code to numerically solve the Lotka-Volterra
model for preditor-prey relationships.

The governing ODEs are
u'(t) = u * (1 - v)
v'(t) = alpha * v * (u - 1)

It serves as a tutorial, with comments along the way.

First a brief note on data types.  The main data type that odeint uses
is numpy.ndarray, which is the data type for NumPy arrays.  The
elements of an ndarray also have a data type, floats by default.
ndarrays are similar to the standard Matlab object (the matrix).  Due
to some nice operator overloading, ndarrays have nice syntax.  For
example, doing the following at a Python prompt will show you some
nice properties.

>>> a = np.array([1.0, 2.0, 3.0, 4.0])
>>> a * a             # Term-by-term multiplication
array([  1.,   4.,   9.,  16.])
>>> np.dot(a, a)      # Vector dot product
30.0
>>> a[2]              # Indexing (indices start at 0)
3.0
>>> a[1:3]            # Slicing (get indexes 1 and 2)
array([ 2.,  3.])
"""


# We have to import modules we want to use:
#  NumPy: numerical packages for Python
#  SciPy: Scientific packages, including odeint; we just import odeint
#  Matplotlib: A library to plot the function; we import it's PyPlot module
import numpy as np
from scipy.integrate import odeint
import matplotlib.pyplot as plt

# The API for ODE is in the SciPy reference guide:
# http://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.odeint.html
# 
# odeint solves a system of ODEs
#   y'(t) = func(y, t)
#   y(0) = y0
#
# y is a vector of values.
#
# odeint expects minimally three arguments in the most basic function call:
#       y = odeint(func, y0, t)
#
# It is convenient to pass parameters into the right-hand-side of our system
# of ODEs, though, so we will also pass the args keyword argument.  Our
# function call is then
#       y = odeint(func, y0, t, args=args)
#
#   func is a function of the form func(y, t).  It must take at least these
#       two arguments, but may have other arguments, such as parameters.
#   y0 is an ndarray containing the initial condition
#   t is an ndarray containing time points you want returned
#   args is a keyword argument that MUST BE A TUPLE containg arguments to
#       be passed to func.
# 
# odeint returns an ndarray y.  y[:,i] will give you the values of the
# ith variable you are solving for at the time points specified by t.


# Ok, let's solve Lotka-Volterra.  We define the L-V function.
def lv_func(y, t, alpha):
    u, v = y  # This is a nice Pythonic way to extract our variables
    return np.array([u * (1.0 - v), 
                     alpha * v * (u - 1.0)])


# Set up L-V solver
def lv_solver(alpha, u_0, v_0, tau):
    """
    A solver for the dimensionless Lotka-Volterra with parameter alpha
    and initial dimensionless prey count of u_0 and predator count of
    v_0.  It returns u and v, the dimensionless prey and predator
    counts sampled at dimensionless time points specified by tau.
    """

    # Construct the initial guess for the dimensionless preditors and prey
    y0 = np.array([u_0, v_0])

    # Call odeint to solve (yes, it is that simple)
    # Note that we pass the parameter alpha as a tuple: (alpha,)
    y = odeint(lv_func, y0, tau, args=(alpha,))

    # Parse to get u and v to return
    return y[:,0], y[:,1]


# Function to plot our result
def plot_lv_result(tau, u, v, pdf_file_name=None):
    """
    Uses Matplotlib to generate plot.  If pdf_file_name is given,
    saves the plot to a PDF.  Otherwise, the plot is displayed on screen.
    """
    plt.close('all')   # Close all windows
    plt.plot(tau, u, 'k-', tau, v, 'b-')  # prey: black line, pred: blue line
    plt.xlabel(r'$\tau$')  # It does \Tex!
    plt.ylabel(r'$u, v$')
    plt.legend(('prey', 'pred'), loc='upper left')  # Legend, upper left
    plt.draw()  # Generate plot
    plt.tight_layout()  # Makes it look nice

    if pdf_file_name is None:
        plt.show()  # This draws it to the screen
    else:
        plt.savefig(pdf_file_name)    
        

# If we're calling from the command line, use alpha = 0.2, u_0 = 10.0,
# v_0 = 2.0, and tau from 0 to 100.
if __name__ == "__main__":
    alpha = 0.2
    u_0 = 10.0
    v_0 = 1.0
    tau = np.linspace(0.0, 100.0, 1000)

    u, v = lv_solver(alpha, u_0, v_0, tau)
    
    plot_lv_result(tau, u, v, pdf_file_name='lv.pdf')


